export interface Url {
  path: string[],
  hash: string,
  search: string
}

/** 
 * use `history.pushState` to push the `path`
 * and dispatch the `popstate` event.
 * 
 * this will trigger the subs rerun.
 */
export function push(path:string): void

/**
 * use `history.replaceState` to replace the `path`
 * and dispatch the `popstate` event.
 * 
 * this will trigger the subs rerun.
 */
export function replace(path:string): void

/**
 * query the global `location` and parse it to
 * `Url`.
 * 
 * warning: you may get old data.
 */
export function getInitUrl(): Url

type WatchCB = (url:Url) => void

export type WatchID = () => void

/**
 * start watch url. When `push` or `replace` called.
 * run the `cb`.
 */
export function watchUrl(cb:WatchCB): WatchID

/**
 * stop watch url. use the given `watchID`.
 */
export function unwatchUrl(watchID:WatchID): void

/**
 * `useUrl` is a React Hook to get `Url` and subscribe
 * the component's render to `watchUrl`.
 * 
 * when the `Url` change(`push` or `replace` get called).
 * the component will rerender to get the fresh data.
 */
export function useUrl(): Url

/**
 * `otherwise` is used to catch all case.
 * 
 * example:
 * matchPath(path, {
 *  [otherwise]: () => <NotFound />
 * })
 */
export const otherwise: unique symbol

type F = (...args:string[]) => unknown

type MatchReturn<
  A extends Record<string|symbol, F>
> = {
  [k in keyof A]: ReturnType<A[k]>
}[keyof A]

/**
 * `matchPath` is a simple helper to match your 
 * url pattern.
 */
export function matchPath<
  A extends Record<string|symbol, F>
>(path: string[], matchOption: A): MatchReturn<A>